(****************************************************************************)
(*                 The Calculus of Inductive Constructions                  *)
(*                                                                          *)
(*                                Projet Coq                                *)
(*                                                                          *)
(*                     INRIA                        ENS-CNRS                *)
(*              Rocquencourt                        Lyon                    *)
(*                                                                          *)
(*                                 Coq V6.1                                 *)
(*                               Oct 1st 1996                               *)
(*                                                                          *)
(****************************************************************************)
(*                                include.ml                                *)
(****************************************************************************)

(* We only assume that the variable COQTOP is set *)
let src_dir =
  let coqtop =
    try Sys.getenv "COQTOP"
    with Not_found -> "." in
  let srccoq = Filename.concat coqtop "src" in
  let src_dir =
    if Sys.file_exists srccoq then srccoq
    else begin
      print_endline ("Cannot find the sources in "^srccoq);
      print_string "Where are they ? ";
      Filename.concat (read_line ()) "src"
    end in
  let add_dir dl = Topdirs.dir_directory
                     (List.fold_left Filename.concat src_dir dl) in
    List.iter add_dir
      [ [ "config" ]; [ "lib"; "util" ]; [ "parsing" ]; [ "meta" ];
        [ "constr" ]; [ "typing" ]; [ "proofs" ]; [ "env" ]; [ "tactics" ] ];
    src_dir
;;
(* Now Coq_config.cmi is accessible *)
Topdirs.dir_directory Coq_config.camlp4lib;;

#use "top_printers.ml";;

#install_printer  (* ast *)  prast;;
#install_printer  (* pat *)  prastpat;;
#install_printer  (* patlist *)  prastpatl;;
#install_printer  (* constr *) ppterm0;;
#install_printer  (* identifier *) prid;;
#install_printer  (* judgement*) prj;;
#install_printer  (* section_path *)  prsp;;
#install_printer  (* goal *)  prgoal;;
#install_printer  (* ctxt *)  prctxt;;
#install_printer  (* proof *)  pproof;;
#install_printer  (* global_constraints *)  prevd;;
#install_printer  (* readable_constraints *)  prevc;;
#install_printer  (* walking_constraints *)  prwc;;
#install_printer  (* constant_object *)  pr_constobj;;
#install_printer  (* universe *)  print_uni;;
#install_printer  (* arc *)  print_arc;;
#install_printer  (* object *)  pp_libobj;;


(* parsing of terms *)

let parse_fterm = Pcoq.parse_string Pcoq.Command.command;;
let parse_ast = Pcoq.parse_string Pcoq.Command.command;;

(* build a term of type constr without type-checking or resolution of 
   implicit syntax *)
let raw_constr_of_string s 
    = Astterm.dbize_cci (Evd.mt_evd()) (unitize_env (gLOB (initial_sign()))) 
      (parse_ast s);;

let e s    = 
   (fconstruct   (Evd.mt_evd()) (initial_sign()) (parse_ast s));;
let infe s = 
   (infconstruct (Evd.mt_evd()) (initial_assumptions()) (parse_ast s));;
(*
let ein gls s = (fconstruct (project gls) (sig_it gls).hyps (parse_ast s));;
let eingl gl s = (fconstruct (Evd.mt_evd()) (sig_it gl).hyps (parse_ast s));;
*)


(* current goal *)

let curgoal = ref (None : goal sigma option);;
let getgoal() =
  try (by (fun g -> curgoal := Some g; failwith "caught") ;failwith "caught")
  with Failure "caught" -> outSOME !curgoal
;;


(* $Id: include.ml,v 1.23 1997/12/19 12:46:30 barras Exp $ *)
