(****************************************************************************)
(*                 The Calculus of Inductive Constructions                  *)
(*                                                                          *)
(*                                Projet Coq                                *)
(*                                                                          *)
(*                     INRIA                        ENS-CNRS                *)
(*              Rocquencourt                        Lyon                    *)
(*                                                                          *)
(*                                 Coq V6.1                                 *)
(*                               Oct 1st 1996                               *)
(*                                                                          *)
(****************************************************************************)
(*                                  Div.v                                   *)
(****************************************************************************)

Require Le.
Require Euclid_def.
Require Compare_dec.

Fixpoint inf_dec [n:nat] : nat->bool :=
      [m:nat] Cases n m of
                  O     _     => true
              | (S n')  O     => false
              | (S n') (S m') => (inf_dec n' m')
             end.



Inductive diveucl [a,b:nat] : Set 
  := divex : (q,r:nat)(gt b r)->(a=(plus (mult q b) r))->(diveucl a b).

Theorem div1 : (b:nat)(gt b O)->(a:nat)(diveucl a b).
Realizer [b,a:nat]Match a with
   (* O *) (O,O)
   (* S *) [n,h]let (q,r) = h in 
              		if (le_gt_dec b (S r)) then
				      ((S q),O)
				 else (q,(S r))
          end.
Program_all.
Rewrite e.
Replace b with (S r).
Simpl.
Elim plus_n_O; Auto.
Apply le_antisym; Auto.
Elim plus_n_Sm; Auto.
Save.

Theorem div2 : (b:nat)(gt b O)->(a:nat)(diveucl a b).
Realizer [b,a:nat]Match a with
   (* O *) (O,O)
   (* S *) [n,h]let (q,r) = h in 
              		 if (inf_dec b (S r)) 
				:: :: { {(le b (S r))}+{(gt b (S r))} } 
				 then ((S q),O)
				 else (q,(S r))
          end.
Program_all.
Rewrite e.
Replace b with (S r).
Simpl.
Elim plus_n_O; Auto.
Apply le_antisym; Auto.
Elim plus_n_Sm; Auto.
Save.


(* $Id: Div.v,v 1.7 1998/01/24 06:14:37 herbelin Exp $ *)
