(****************************************************************************)
(*                 The Calculus of Inductive Constructions                  *)
(*                                                                          *)
(*                                Projet Coq                                *)
(*                                                                          *)
(*                     INRIA                        ENS-CNRS                *)
(*              Rocquencourt                        Lyon                    *)
(*                                                                          *)
(*                                 Coq V6.1                                 *)
(*                               Oct 1st 1996                               *)
(*                                                                          *)
(****************************************************************************)
(*                              LogicSyntax.v                               *)
(****************************************************************************)

(* Parsing of things in Logic.v *)

Grammar command command1 :=
  conj [ "<" lcommand($l1) "," lcommand($c2) ">" "{" command($c3) ","
        command($c4) "}" ] -> [<<(conj $l1 $c2 $c3 $c4)>>]
| fst [ "<" lcommand($l1) "," lcommand($c2) ">" "Fst" "{"
        lcommand($l) "}" ] -> [<<(proj1 $l1 $c2 $l)>>]
| snd [ "<" lcommand($l1) "," lcommand($c2) ">" "Snd" "{"
        lcommand($l) "}" ] -> [<<(proj2 $l1 $c2 $l)>>]
| IF [ "either" command($c) "and_then" command($t) "or_else" command($e) ] ->
       [<<(IF $c $t $e)>>]
| all [ "<" lcommand($l1) ">" "All" "(" lcommand($l2) ")" ] ->
       [<<(all $l1 $l2)>>]
| eq_expl [ "<" lcommand($l1) ">" command0($c1) "=" command0($c2) ] ->
            [<<(eq $l1 $c1 $c2)>>]
| eq_impl [ command0($c) "=" command0($c2) ] -> [<<(eq ? $c $c2)>>]

with command2 :=
  not [ "~" command2($c) ] -> [<<(not $c)>>]

with command6 :=
  and [ command5($c1) "/\\" command6($c2) ] -> [<<(and $c1 $c2)>>]

with command7 :=
  or [ command6($c1) "\\/" command7($c2) ] -> [<<(or $c1 $c2)>>]

with command8 :=
  iff [ command7($c1) "<->" command8($c2) ] -> [<<(iff $c1 $c2)>>]

with command10 :=
  allexplicit [ "ALL" ident($x) ":" command($t) "|" command($p) ]
                          -> [<<(all $t [$x:$t]$p)>>]
| allimplicit [ "ALL" ident($x) "|" command($p) ]
                          -> [<<(all ? [$x]$p)>>]
| exexplicit [ "EX" ident($v) ":" command($t) "|" command($c1) ]
                          -> [<<(ex $t [$v:$t]$c1)>>]
| eximplicit [ "EX" ident($v) "|" command($c1) ] 
                          -> [<<(ex ? [$v]$c1)>>]
| ex2explicit [ "EX" ident($v) ":" command($t) "|" command($c1) "&"
           command($c2) ] -> [<<(ex2 $t [$v:$t]$c1 [$v:$t]$c2)>>]
| ex2implicit [ "EX" ident($v) "|" command($c1) "&" 
           command($c2) ] -> [<<(ex2 ? [$v]$c1 [$v]$c2)>>].

(* $Id: LogicSyntax.v,v 1.10 1997/11/28 14:00:52 barras Exp $ *)
